// Solver.cpp: implementation of the Solver class.
//
//////////////////////////////////////////////////////////////////////

#ifdef _MSC_VER
// Disable the warning n. 4786
#pragma warning( disable : 4786 )
#endif

#include <vector>

#include "Solver.h"

namespace ucf
{

	using namespace oti;
	using namespace std;

//////////////////////////////////////////////////////////////////////
// class Solver
//////////////////////////////////////////////////////////////////////

Solver::tOptEnvironment* Solver::fOptEnvironment=NULL;
int Solver::fOptEnvUsers=0;

//////////////////////////////////////////////////////////////////////
//
//////////////////////////////////////////////////////////////////////

Solver::Solver()
	:fObjectiveFunction(NULL)
	,fConstraintsSet(0)
	,fOptProblem(NULL)
{

}

//////////////////////////////////////////////////////////////////////
//
//////////////////////////////////////////////////////////////////////

Solver::~Solver()
{
   if (fOptProblem==NULL)
      return;

	delete fOptProblem;
	--fOptEnvUsers;

	if (fOptEnvUsers==0)
		delete fOptEnvironment;
}

//////////////////////////////////////////////////////////////////////
//
//////////////////////////////////////////////////////////////////////

void Solver::StartInitialization(
		Variables& variables,
		ObjectiveFunction& objectiveFunction)
{
	// setup the oti solver
	assert ( (fOptEnvironment==NULL && fOptEnvUsers==0) ||
				(fOptEnvironment!=NULL && fOptEnvUsers>0));

	if (fOptEnvUsers==0)
		fOptEnvironment = new tOptEnvironment;

	fOptProblem = new tOptProblem(*fOptEnvironment);
	++fOptEnvUsers;

	// Registers variables and objective function.
	assert (fRegisteredVariables.fVariables==NULL);
	assert (fObjectiveFunction==NULL);
	assert (fConstraintsSet.size()==0);

	fRegisteredVariables.fVariables = &variables;
	fRegisteredVariables.fActorID = variables.RegisterActor();

	fObjectiveFunction = &objectiveFunction;
	fOptProblem->chgobjsen(
		fObjectiveFunction->GetMinimization()
			?OtiStatic::MINIMIZE
			:OtiStatic::MAXIMIZE
			);

	// Initialize the fOptProblem columns.
	this->InitializeColumns();
}

//////////////////////////////////////////////////////////////////////
//
//////////////////////////////////////////////////////////////////////

void Solver::InitializeColumns()
{
	if (fObjectiveFunction->GetFunctionType()==ObjectiveFunction::eLinear)
		fOptProblem->chgprobtype(tOptProblem::eLp);
	else if (fObjectiveFunction->GetFunctionType()==ObjectiveFunction::eQuadratic)
		fOptProblem->chgprobtype(tOptProblem::eQp);
	else assert (false);

	const Variables* variables = fRegisteredVariables.fVariables;
	if (fObjectiveFunction->GetFunctionType()>=ObjectiveFunction::eLinear)
	{
		// set the linear terms of objective function
		const int ccnt = variables->GetNumVars();
		vector<double> obj, lb, ub;
		vector<OtiStatic::eVarType> vtype;

		lb.reserve(ccnt);
		ub.reserve(ccnt);
		vtype.reserve(ccnt);

		obj.reserve(ccnt);
		ObjectiveFunction::tVector sparseObj;
		fObjectiveFunction->GetLinearCoeffs(sparseObj);
		sparseObj.Densify(obj.begin());

		const double kInfinite = fOptProblem->getinfinite();
		for (Variables::tIndex group=0; group<variables->GetNumGroups(); ++group)
		{
			for (Variables::tIndex i=0; i<variables->GetCardGroup(group); ++i)
			{
				int idVar = variables->ToGlobalIndex(group, i);
				assert (idVar<ccnt);
				lb[idVar] = variables->GetLowerBound(group, i);
				if (lb[idVar]<-kInfinite)
					lb[idVar]=-kInfinite;
				ub[idVar] = variables->GetUpperBound(group, i);
				if (ub[idVar]>kInfinite)
					ub[idVar]=kInfinite;
				vtype[idVar] = variables->GetVarTypeGroup(group);
			}
		}

		fOptProblem->newcols(ccnt, obj.begin(),
			lb.begin(), ub.begin(), vtype.begin());
	}

	if (fObjectiveFunction->GetFunctionType()>=ObjectiveFunction::eQuadratic)
	{
		// set the quadratic terms of objective function
		const int ccnt = variables->GetNumVars();
		for (Variables::tIndex i=0; i<ccnt; ++i)
		{
			ObjectiveFunction::tVector coeffs;
			fObjectiveFunction->GetQuadCoeffs(coeffs, i);
			for (int j=0; j<coeffs.GetBaseDim(); ++j)
			{
				fOptProblem->chgqpcoef(i, coeffs.Base(j), coeffs.NonZero(j));
			}
		}
	}

	if (fObjectiveFunction->GetFunctionType()>ObjectiveFunction::eQuadratic)
	{
		assert (false);
	}

}

//////////////////////////////////////////////////////////////////////
//
//////////////////////////////////////////////////////////////////////

void Solver::AddConstraints(Constraints& constraints)
{
	// Register as actor
	tConstraintsSet::iterator lastConstr;
	if (fConstraintsSet.size()==0)
	{
		fConstraintsSet.insert(fConstraintsSet.end(), RegisteredConstraints());
		lastConstr = fConstraintsSet.begin();
	}
	else
	{
		lastConstr = fConstraintsSet.end();
		fConstraintsSet.insert(fConstraintsSet.end(), RegisteredConstraints());
	}
	lastConstr->fConstraints = &constraints;
	lastConstr->fActorID = constraints.RegisterActor();
	
	// Build the constraints in the fOptProblem
	assert (fRegisteredVariables.fVariables == constraints.GetVariables());
	const int rcnt = constraints.GetNumRows();

	for (int row=0; row<rcnt; ++row)
	{
		Constraints::tSparseVector coefficients;
		Constraints::eConstrType constrType;
		double rhs, rngval;
		constraints.GetConstraint(row, coefficients, constrType, rhs, rngval);
		const int nzcnt = coefficients.GetBaseDim();
		int rmatbeg[2] = {0, nzcnt};

		fOptProblem->addrows(
			0, 1, nzcnt, &rhs, &constrType,
			rmatbeg, coefficients.Base(), coefficients.NonZero());
			
		if (constrType == OtiStatic::eRanged)
		{
			assert(rngval>0);
			int row = fOptProblem->numrows();
			fOptProblem->chgrngval(1, &row, &rngval);
		}
		
		// Register the row name
		constraints.SetConstraintGlobalIndex(row, fOptProblem->numrows());
	}

}

//////////////////////////////////////////////////////////////////////
//
//////////////////////////////////////////////////////////////////////

void Solver::EndInitialization()
{
	// Check that at least one constraint was inserted.
	assert (fConstraintsSet.size()>0);
}

//////////////////////////////////////////////////////////////////////
//
//////////////////////////////////////////////////////////////////////

void Solver::Solve()
{

	this->CheckAndUpdateVariables();
	this->CheckAndUpdateConstraints();

	if (fObjectiveFunction->GetFunctionType()==ObjectiveFunction::eLinear)
	{
		fOptProblem->write("problema.lp");
		fOptProblem->dualopt();
		return;
	}
	if (fObjectiveFunction->GetFunctionType()==ObjectiveFunction::eQuadratic)
	{
		fOptProblem->write("problema.lp");
		fOptProblem->baropt();
		return;
	}

	assert (false);
}

//////////////////////////////////////////////////////////////////////
//
//////////////////////////////////////////////////////////////////////

void Solver::CheckAndUpdateVariables()
{
	Variables::tModifications* modification = NULL;
	fRegisteredVariables.fVariables->GetModification(
		fRegisteredVariables.fActorID, modification);

	assert (modification!=NULL);
	Variables::tModifications::iterator iterModification = modification->begin();
	while (iterModification != modification->end())
	{
		Variables::Modification* modification = *iterModification;
		switch (modification->fTagModification)
		{
		case Variables::eSetVar:
		{
			Variables::FixSetVarMod* fixSetVarMod =
				static_cast<Variables::FixSetVarMod*>(modification);
			this->FixSetVar(fixSetVarMod);
			break;
		}
		case Variables::eFixedVar:
		{
			Variables::FixSetVarMod* fixSetVarMod =
				static_cast<Variables::FixSetVarMod*>(modification);
			this->FixSetVar(fixSetVarMod);
			break;
		}
		case Variables::eFreedVar:
		{
			Variables::UnfixVarMod* unfixVarMod =
				static_cast<Variables::UnfixVarMod*>(modification);
			this->FreedVar(unfixVarMod);
			break;
		}
		default:
			assert (false);
		}

		iterModification++;
	}

	Variables::FreeModifications(modification);
}

//////////////////////////////////////////////////////////////////////
//
//////////////////////////////////////////////////////////////////////

void Solver::FixSetVar(const Variables::FixSetVarMod* fixSetVarMod)
{
	Variables::tIndex groupName = fixSetVarMod->fGroupName;
	Variables::tIndex varName = fixSetVarMod->fVarName;
	int index = fRegisteredVariables.fVariables->
		ToGlobalIndex(groupName, varName);
	char lu='B';
	double bound = fixSetVarMod->fValue;

	fOptProblem->chgbds(1, &index, &lu, &bound);
}

//////////////////////////////////////////////////////////////////////
//
//////////////////////////////////////////////////////////////////////

void Solver::FreedVar(const Variables::UnfixVarMod* unfixVarMod)
{
	Variables::tIndex groupName = unfixVarMod->fGroupName;
	Variables::tIndex varName = unfixVarMod->fVarName;
	int index = fRegisteredVariables.fVariables->
		ToGlobalIndex(groupName, varName);
	char lu='L';
	double bound = fRegisteredVariables.fVariables->
		GetLowerBound(groupName, varName);
	fOptProblem->chgbds(1, &index, &lu, &bound);
	
	lu='U';
	bound = fRegisteredVariables.fVariables->
		GetUpperBound(groupName, varName);
	fOptProblem->chgbds(1, &index, &lu, &bound);
}

//////////////////////////////////////////////////////////////////////
//
//////////////////////////////////////////////////////////////////////

void Solver::CheckAndUpdateConstraints()
{
	// Iter the registered constraints
	tConstraintsSet::iterator iterConstr = fConstraintsSet.begin();
	while (iterConstr!=fConstraintsSet.end())
	{
		const RegisteredConstraints& registeredConstraints = *iterConstr;

		// Iter the modification
		Constraints::tModifications* modifications = NULL;
		registeredConstraints.fConstraints->GetModification(
			registeredConstraints.fActorID, modifications);
		assert (modifications != NULL);
		Constraints::tModifications::iterator iterModification = modifications->begin();
		while (iterModification != modifications->end())
		{
			// Actually no constraints modification
			assert (false);

			++iterModification;
		}

		Constraints::FreeModifications(modifications);
		iterConstr++;
	}
}

} // namespace

