// Solver.h: interface for the Solver class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_SOLVER_H)
#define AFX_SOLVER_H

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include <deque>

#include "Variables.h"
#include "ObjectiveFunction.h"
#include "Constraints.h"
#include "oti.h"

namespace ucf
{

class Solver  
{
public:
#ifdef __USE_OTI_OSL__
	typedef oti::OptEnvironment<oti::eOsl> tOptEnvironment;
	typedef oti::OptProblem<oti::eOsl> tOptProblem;
#endif
#ifdef __USE_OTI_CPX__
	typedef oti::OptEnvironment<oti::eCplex> tOptEnvironment;
	typedef oti::OptProblem<oti::eCplex> tOptProblem;
#endif

	Solver();

	virtual ~Solver();

	/*
	Start the initialization of solver, buildin the internal
	status of Solver and allocating the columns of fOptProblem.
	If necessary, also the fOptEnvironment will be initialized.
	*/
	void StartInitialization(
		Variables& variables,
		ObjectiveFunction& objectiveFunction);
	/*
	Add one set of constraints to fOptProblem allocating the rows of
	fOptProblem.
	*/
	void AddConstraints(Constraints& constraints);
	/*
	Terminate the initialization of solver, that now is ready to solve
	the problem.
	*/
	void EndInitialization();

	/*
	Solve() call the solver on the problem in memory.
	The status of computation can be read with GetStatus().
	The value of optimum can be read with GetObjectiveVal().
	The primals and duals values can be read whit GetPrimals() and
	GetDuals() respectively.
	*/
	void Solve();
	typedef tOptProblem::eOTIReturnCode eStatus;
	eStatus GetStatus() const { return fOptProblem->otistat(); }
	double GetObjectiveVal() const { return fOptProblem->objval(); }
	void GetPrimals(double primals[]) const { fOptProblem->getX(primals); }
	void GetDuals(double duals[]) const { fOptProblem->getPi(duals); }

private:
	struct RegisteredVariables
	{
		RegisteredVariables()
			:fVariables(NULL)
			,fActorID(-1)
		{};

		const Variables* fVariables;
		Variables::tIndex fActorID;
	};

	struct RegisteredConstraints
	{
		RegisteredConstraints()
			:fConstraints(NULL)
			,fActorID(-1)
		{};

		const Constraints* fConstraints;
		Variables::tIndex fActorID;
	};
	typedef std::deque<RegisteredConstraints> tConstraintsSet;


	/*
	The following filds describe the problem to be solved.
	*/
	RegisteredVariables fRegisteredVariables;
	const ObjectiveFunction* fObjectiveFunction;
	tConstraintsSet fConstraintsSet;

	/*
	The solver.
	*/


	static tOptEnvironment* fOptEnvironment;
	static int fOptEnvUsers;
	tOptProblem* fOptProblem;

	/*
	Initialize the columns of fOptProblem.
	*/
	void InitializeColumns();

	/*
	Check if variables are changed and modify the
	fOptProblem consequently.
	*/
	void CheckAndUpdateVariables();

	/*
	Impose the value of the variable (fixSetVarMod->fGroupName, 
	fixSetVarMod->fVarName) at fixSetVarMod->fValue.
	As solver of linear and quadratic problem, we don't make difference
	between fixed and set variable.
	*/
	void FixSetVar(const Variables::FixSetVarMod* fixSetVarMod);

	/*
	Restore the original constraint of variable (unfixVarMod->fGroupName,
	unfixVarMod->fVarName).
	*/
	void FreedVar(const Variables::UnfixVarMod* unfixVarMod);

	/*
	Check if constraint are changed and modify the
	fOptProblem consequently.
	*/
	void CheckAndUpdateConstraints();
};

} // namespace

#endif // !defined(AFX_SOLVER_H)
