// Variables.h: interface for the Variables class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_VARIABLES_H)
#define AFX_VARIABLES_H

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include <cassert>
// #include <limits>
#include <otilimits.h>
#include <vector>
#include <list>

#include "oti.h"
#include "ActorModification.h"

namespace ucf
{

class Group; // Forward

class Variables  
{
public:
	typedef int tIndex;
	typedef double tVarValue;
	typedef std::vector<tIndex> tGroupsName;

	static const tVarValue kInfinite; //= std::numeric_limits<tVarValue>::infinity();
	typedef oti::OtiStatic::eVarType eVarType;

	Variables();
	virtual ~Variables();

	void RegisterVars(eVarType type, tIndex numOfVar, tIndex& groupName);
	void SetVarRange(tIndex groupName, tIndex varName,
		tVarValue lowerBound, tVarValue upperBound);
	void EndRegistration();

	tIndex GetNumVars() const { return fNumVars; }
	tIndex GetNumGroups() const { return fFirstFreeGroup; }
	tIndex GetCardGroup(tIndex groupName) const;
	eVarType GetVarTypeGroup(tIndex groupName) const;
	void GetVarRange(tIndex groupName, tIndex varName,
		tVarValue& lowerBound, tVarValue& upperBound) const;
	tVarValue GetLowerBound(tIndex groupName, tIndex varName) const;
	tVarValue GetUpperBound(tIndex groupName, tIndex varName) const;

	tIndex ToGlobalIndex(tIndex groupName, tIndex varName) const;
	void ToLocalIndex(tIndex globalName, tIndex& groupName,
		tIndex& varName) const;

	void SetVar(tIndex groupName, tIndex varName, tVarValue value);
	void FixVar(tIndex groupName, tIndex varName, tVarValue value);
	void FreeVar(tIndex groupName, tIndex varName);
	
	enum eModification
	{
		eSetVar = 0,
		eFixedVar,
		eFreedVar
	};

	class Modification
	{
	public:
		eModification fTagModification;

		Modification(eModification modification)
			:fTagModification(modification)
		{}

		virtual ~Modification() {};

	};
	class FixSetVarMod : public Modification
	{
	public:
		tIndex fGroupName;
		tIndex fVarName;
		tVarValue fValue;

		FixSetVarMod(eModification modification, tIndex groupName,
			tIndex varName, tVarValue value)
			:Modification(modification)
			,fGroupName(groupName)
			,fVarName(varName)
			,fValue(value)
		{}

		virtual ~FixSetVarMod() {};

	};
	class UnfixVarMod : public Modification
	{
	public:
		tIndex fGroupName;
		tIndex fVarName;

		UnfixVarMod(eModification modification, tIndex groupName,
			tIndex varName)
			:Modification(modification)
			,fGroupName(groupName)
			,fVarName(varName)
		{}

		virtual ~UnfixVarMod() {};

	};

	typedef ActorModification<Modification> tActorModification;
	typedef tActorModification::tModifications tModifications;

	tIndex RegisterActor();
	void GetModification(tIndex actor, tModifications*& modifications) const;
	static void FreeModifications(tModifications* modifications);

	/*
	Retrive the status of variables.
	Warning: when GetVarStatus(g, i)==eFreedVar the behavior of
	GetVarValue(g, i) is undefined!
	*/
	eModification GetVarStatus(tIndex groupName, tIndex varName) const;
	tVarValue GetVarValue(tIndex groupName, tIndex varName) const;

private:
	typedef std::vector<Group*> tVarsGroups;
	typedef std::vector<tIndex> tGlobalNames;
	typedef std::vector<tActorModification*> tVarActors;

	tIndex fFirstFreeGroup;
	tVarsGroups fVarsGroups;
	tGlobalNames fGlobalNames;

	tIndex fFirstFreeActor;
	tVarActors fVarActors;

	tIndex fNumVars;
};

/*
Describe a group of variables all of the same type. Every call to 
Variables::RegisterVars() build a new group.
*/
class Group
{
public:
	typedef Variables::tIndex tIndex;
	typedef Variables::tVarValue tVarValue;
	typedef Variables::eVarType eVarType;
	typedef Variables::eModification eVarStatus; // The possible status
		// of a variable are isomorphe to the set of possible
		// modification for a variable.

	/*
	Construct a Group of numVars variables of type varType. Default
	values for lower and upper bound are -kInfinite and kInfinite,
	respectively.
	*/
	Group(eVarType varType, tIndex numVars);

	/*
	Modify the lower and upper bound for variable varId.
	*/
	void ModifyLowerBound(tIndex varId, tVarValue lowerBound);
	void ModifyUpperBound(tIndex varId, tVarValue upperBound);

	/*
	Respectively, set, fix and free the value of a variable.
	For more detayls on semanthic of set, fix and free, please refear to
	class Variables documentation.
	*/
	void SetVar(tIndex varName, tVarValue value);
	void FixVar(tIndex varName, tVarValue value);
	void FreeVar(tIndex varName);
	
	/*
	Retrive the status of variables.
	Warning: when GetVarStatus(i)==eFreedVar the behavior of
	GetVarValue(i) is undefined!
	*/
	eVarStatus GetVarStatus(tIndex varName) const { return fVarsStatus[varName]; }
	tVarValue GetVarValue(tIndex varName) const { return fVarsValues[varName]; }

	/*
	Returns the lower and upper bound for variable varId.
	*/
	tVarValue GetLowerBound(tIndex varId) { return fLowerBounds[varId]; }
	tVarValue GetUpperBound(tIndex varId) { return fUpperBounds[varId]; }

	/*
	Return the number of variables in the group.
	*/
	tIndex Cardinality() const { return fNumVars; }

	/*
	Return the type of variables in the group.
	*/
	eVarType VarType() const { return fVarType; }

private:
	typedef std::vector<tVarValue> tVarsValues;
	typedef std::vector<tVarValue> tVarsBounds;
	typedef std::vector<eVarStatus> tVarsStatus;

	eVarType fVarType;
	tVarsStatus fVarsStatus;
	tVarsValues fVarsValues;
	tIndex fNumVars;
	tVarsBounds fLowerBounds;
	tVarsBounds fUpperBounds;
};


//////////////////////////////////////////////////////////////////////
// inline function
//////////////////////////////////////////////////////////////////////

inline Variables::tIndex Variables::GetCardGroup(tIndex groupName) const
{
	return fVarsGroups[groupName]->Cardinality();
}

inline Variables::eVarType Variables::GetVarTypeGroup(tIndex groupName) const
{
	return fVarsGroups[groupName]->VarType();
}

inline void Variables::GetVarRange(tIndex groupName, tIndex varName,
		tVarValue& lowerBound, tVarValue& upperBound) const
{
	Group* group = fVarsGroups[groupName];
	lowerBound = group->GetLowerBound(varName);
	upperBound = group->GetUpperBound(varName);
}

inline Variables::tVarValue Variables::GetLowerBound(tIndex groupName, tIndex varName) const
{
	return fVarsGroups[groupName]->GetLowerBound(varName);
}

inline Variables::tVarValue Variables::GetUpperBound(tIndex groupName, tIndex varName) const
{
	return fVarsGroups[groupName]->GetUpperBound(varName);
}

inline Variables::tIndex Variables::ToGlobalIndex(tIndex groupName, tIndex varName) const
{
	return fGlobalNames[groupName] + varName;
}

inline void Variables::GetModification(tIndex actor, tModifications*& modifications) const
{
	assert (actor>=0 && actor<tIndex(fVarActors.size()));
	fVarActors[actor]->GetModification(modifications);
}

inline Variables::eModification Variables::GetVarStatus(tIndex groupName, tIndex varName) const
{
	return fVarsGroups[groupName]->GetVarStatus(varName);
}

inline Variables::tVarValue Variables::GetVarValue(tIndex groupName, tIndex varName) const
{
	return fVarsGroups[groupName]->GetVarValue(varName);
}

//////////////////////////////////////////////////////////////////////
//
//////////////////////////////////////////////////////////////////////

inline Group::Group(eVarType varType, tIndex numVars)
	:fVarType(varType)
	,fVarsStatus(numVars, Variables::eFreedVar)
	,fVarsValues(numVars)
	,fNumVars(numVars)
	,fLowerBounds(numVars, -Variables::kInfinite)
	,fUpperBounds(numVars, Variables::kInfinite)
{
}

inline void Group::ModifyLowerBound(tIndex varId, tVarValue lowerBound)
{
	fLowerBounds[varId] = lowerBound;
}

inline void Group::ModifyUpperBound(tIndex varId, tVarValue upperBound)
{
	fUpperBounds[varId] = upperBound;
}

inline void Group::SetVar(tIndex varName, tVarValue value)
{
	assert (fLowerBounds[varName] <= value);
	assert (value <= fUpperBounds[varName]);
	fVarsStatus[varName] = Variables::eSetVar;
	fVarsValues[varName] = value;
}

inline void Group::FixVar(tIndex varName, tVarValue value)
{
	assert (fLowerBounds[varName] <= value);
	assert (value <= fUpperBounds[varName]);
	fVarsStatus[varName] = Variables::eFixedVar;
	fVarsValues[varName] = value;
}

inline void Group::FreeVar(tIndex varName)
{
	fVarsStatus[varName] = Variables::eFreedVar;
}

} // namespace

#endif // !defined(AFX_VARIABLES_H)
